/*
 * $Header: /sources/ces/cesOsApi/common/src/cesResourceTable.c,v 1.43 2009/02/02 14:22:22 andreas Exp $
 *
 * $Log: cesResourceTable.c,v $
 * Revision 1.43  2009/02/02 14:22:22  andreas
 * - add first draft of resource table for MFCC 8463
 *
 * Revision 1.42  2008/08/05 13:25:27  ricco
 * Adapted to run on VxWorks 653 partition (VTHREADS).
 *
 * Revision 1.41  2008/07/29 18:01:52  rupert
 * Map all resources to CPU A on single-pro RIO4s
 *
 * Revision 1.40  2008/06/12 15:09:48  andreas
 * - disable update from PPCMon table for
 *  - serial flash
 *  - fram
 *   - RTC
 * Those are known to contain incorrect entries on some boards.
 *
 * Revision 1.39  2008/03/13 20:10:46  rupert
 * Properly detect 8064WD
 *
 * Revision 1.38  2008/03/10 20:07:05  rupert
 * - Update a lot of resources for the 4070
 * - Add CESRT_LOCAL_CSR to documentation and resource tables
 * - Add a 3rd version number to serial chips
 * - Improve debugging in cesResourceTable.c
 * - Implement irq non-rotation of PMC#2 slot on some
 *   types of rio3
 * - Update documentation
 *
 * Revision 1.37  2007/12/05 07:58:12  andreas
 * - do not init from PPCMon table on 4070 and similar boards because this will override the NVRAM/RTC base address. On those boards the resource table should indicate the base address used to perform register based access and not the memory mapped base address.
 *
 * Revision 1.36  2007/10/18 20:13:44  rupert
 * Remove all unused resources
 *
 * Revision 1.35  2007/10/17 14:59:08  rupert.rupt
 * Properly detect all RIO4 types
 *
 * Revision 1.34  2007/10/10 13:02:31  andreas
 * - remove unused variable
 *
 * Revision 1.33  2007/08/06 20:31:33  rupert
 * Don't read the ResourceTable for now. The resource table
 * will not be read from FRAM in this release yet.
 *
 * Revision 1.32  2007/08/06 20:27:54  rupert
 * Better detect fram/nvram:
 * - if board type is RIO4, RIOS or RIOC4070Style, we try to read FRAM
 * - else, we try to read NVRAM
 *
 * Revision 1.31  2007/08/06 20:24:35  rupert
 * Init CESRT_BOARD_BASE with board type and serial number
 *
 * Revision 1.30  2007/08/06 20:11:06  rupert
 * Move the resource tables into separate files
 *
 * Revision 1.29  2007/08/06 20:10:09  rupert
 * Read the resource table from FRAM
 *
 * Revision 1.28  2007/08/06 20:09:13  rupert
 * add macro BTYPE_IS_RIO4
 *
 * Revision 1.27  2007/08/06 20:05:11  rupert
 * Add resource table for RIOC
 *
 * Revision 1.26  2007/06/07 15:58:54  andreas
 * - add resource table setup for 8071
 *
 * Revision 1.25  2007/05/04 13:30:52  rupert
 * - Add entries for PCI resources
 * - Add a table for RIO-S
 *
 * Revision 1.24  2006/11/18 10:21:31  akos
 * Add FIFO reads to resource table on RIO4
 *
 * Revision 1.23  2006/10/18 12:06:05  rupert
 * Update table dump
 *
 * Revision 1.22  2006/10/17 14:00:57  akos
 * Add const to argument of cesRT_GetPciSlot
 *
 * Revision 1.21  2006/09/29 14:25:46  akos
 * Cosmetics
 *
 * Revision 1.20  2006/07/19 14:07:40  akos
 * 03-table_update.diff
 * Updates to the resource table:
 * - present resources must not have address = 0xffffffff, because
 *   this indicates "resource not present"
 *
 * Revision 1.19  2006/07/17 16:28:30  andreas
 * - fix compiler warnings for GHS compiler
 *
 * Revision 1.18  2006/05/10 14:20:26  andreas
 * - add FRAM/RTC resources to 4070 configuration
 *
 * Revision 1.17  2006/05/08 13:59:03  akos
 * patches for sugarhat release 3.0.0-alpha4
 *
 * Revision 1.16  2006/05/02 12:44:37  rupert
 * fix compiler errors and warnings
 *
 * Revision 1.15  2006/05/02 12:07:34  rupert
 * improve initialization for rio4 case
 *
 * Revision 1.14  2006/04/20 14:38:03  andreas
 * - diab cannot handle __typeof
 *
 * Revision 1.13  2006/03/21 17:08:52  andreas
 * - add table for classic MFCCs
 *
 * Revision 1.12  2006/03/16 10:48:14  andreas
 * - add initialization for 4070 and compatible
 * - add some missing base addresses for flash stuff
 *
 * Revision 1.11  2006/02/20 15:42:17  andreas
 * - get rid of C++ comments
 *
 * Revision 1.10  2006/02/20 13:35:06  akos
 * (fixed conflicts in common/src/cesResourceTable.c)
 * cesOsApi/01-linux_boot_from_table.diff
 * The code is now able to create tables which work on both processors of my
 * RIO4 and on my RIO3.
 * - remove now unneeded code from Linux/src/cesResourceTableInit.c. the filename
 *   is now misleading, but maybe the file will disappear completely in future.
 *
 * Revision 1.9  2006/02/10 08:47:23  bettaieb
 * minor updates for compilation
 *
 * Revision 1.8  2006/02/03 13:33:52  rupert
 * more initialization code
 *
 * Revision 1.7  2006/01/26 14:23:56  akos
 * Fix compilation problems on Linux
 *
 * Revision 1.6  2006/01/26 09:22:43  akos
 * Implement some initialization functions
 *
 */

#ifndef NULL
#define NULL ((void*)0L)
#endif

#include "ces/cesResourceTable.h"
#include "ces/cesResourceTableP.h"
#include "ces/cesBoardType.h"
#include "ces/cesMalloc.h"
#include "ces/cesPrintf.h"

#undef DEBUG
#ifdef DEBUG
#if 1
extern void sercon_putc(char);
extern void sercon_puts(char*);
static void sercon_puthex(unsigned int number) {
	int i;
	int n;
	for (i = 0; i < 8; i++) {
		n = number >> 28;
		if (n < 10)
			sercon_putc('0'+n);
		else
			sercon_putc('a'-10+n);
		number <<= 4;
	}
}
#define DBG_C    sercon_putc
#define DBG_S(s) sercon_puts(PTR_OFFSET(s))
#define DBG_H    sercon_puthex
#else
extern int printf(const char *fmt, ...);
#define DBG(fmt, x...) printf(PTR_OFFSET(fmt), ##x)
#define DBG_C(c) DBG("%c", c)
#define DBG_S(s) DBG("%s", s)
#define DBG_H(v) DBG("%x", v)
#endif

#else
#define DBG(x...) do{}while(0)
#define DBG_C(c)
#define DBG_S(s)
#define DBG_H(v)
#endif


#define CESRT_MAGIC 0xce547000
#define FPROM_PPCMON_VERSION(fprom_addr) (*(long*)( (unsigned long)(fprom_addr) + 0x0c ))
#define FPROM_BOARD_TYPE(fprom_addr)     (*(long*)( (unsigned long)(fprom_addr) + 0x1c ))
#define FPROM_NVRAM_ADDR(fprom_addr)     (*(long*)( (unsigned long)(fprom_addr) + 0x54 ))
#define NVRAM_BOARD_NAME(nvram_addr, i)  (*(char*)(nvram_addr + 0x1007 + i * 4))
#define CES4070 0x34303730

#define BTYPE_IS_RIO4(b)  ((b & 0xffffff00) == 0x38303700)

/* is rio3 if
  - 806x && x >= 4
    or
  - 406x && x >= 5
*/
#define BTYPE_IS_RIO3_VME(b) (((b & 0xffffff00) == 0x38303600 && b >= 0x38303634))
#define BTYPE_IS_RIO3_CPCI(b) (((b & 0xffffff00) == 0x34303600 && b >= 0x34303635))

#define BTYPE_IS_RIO3(b) (BTYPE_IS_RIO3_VME(b) || BTYPE_IS_RIO3_CPCI(b))


/* is 4070style if
  4070
  7448
  8448
  8449
  7449
*/
#define BTYPE_IS_RIOC4070STYLE(b) ( (b == 0x34303730) || (b == 0x37343438) || \
                                   (b == 0x38343438) || (b == 0x38343439) || (b == 0x37343439) )
/*
  is classic MFCC if:
  8441
  8442
  8443
  8444
  8445
  8446
  8447
*/
#define BTYPE_IS_CLASSICMFCC(b) ( (b == 0x38343431) || (b == 0x38343432) || \
                                  (b == 0x38343433) || (b == 0x38343434) || \
                                  (b == 0x38343435) || (b == 0x38343436) || \
                                  (b == 0x38343437) )
/* is RIOS if:
   2476
*/
#define BTYPE_IS_RIOS(b)        (b == 0x32343736)

/* 8463 style MFCC */
#define BTYPE_IS_8463STYLEMFCC(b) ( b == 0x38343633 )

/* These macros require special implicit arguments */
#ifdef __GNUC__
#define PTR_OFFSET(x)     ((__typeof(&(x)[0]))((unsigned long)(&(x)[0]) + offset))
#else
/* Windriver Compiler does not interpret __typeof. Since this macro is
only necessary on Linux let us just define it empty if a non-GNU compiler is used */
#define PTR_OFFSET(x)     (x)
#endif
#define PHYS_TO_VIRT(x)  (phys_to_virt ? phys_to_virt(x) : (void*)(x))

/* WARNING: the table order is different from the resource list!! */
typedef struct {
  CesResourceID id;
  CesResourceTableEntry_t entry;
} CesResourceDescriptor_t;

#define table rio3Resources
#include "rio3Resources.cesRT"
#undef table

#define table rio3CpciResources
#include "rio3CpciResources.cesRT"
#undef table

#define table rio3VmeResources
#include "rio3VmeResources.cesRT"
#undef table

#define table rioc4070styleResources
#include "rioc4070styleResources.cesRT"
#undef table

#define table rioc4070Resources
#include "rioc4070Resources.cesRT"
#undef table

#define table classicMfccResources
#include "classicMfccResources.cesRT"
#undef table

#define table rio4Resources
#include "rio4Resources.cesRT"
#undef table

#define table rioSResources
#include "rioSResources.cesRT"
#undef table

#define table mfcc8463styleResources
#include "mfcc8463styleResources.cesRT"
#undef table

CesResourceDescriptor_t rio3NoIrqRotationResources[] =
{                        /* address      size     version   flags  extra data */
  /* Represents the non-rotated irqs on RIOC, RIO3 8064W and the low-cost RIOs. */
  {CESRT_LPCI_PMC2,       { 0x15171819, 0x00000000, 0x0000, 0x0001, 0x00000000}}, 
  {CESRT_INVALID_RESOURCE,{ 0xffffffff, 0x00000000, 0x0000, 0x0000, 0x00000000}}
};

/* this is a table of the resources that match the PPCMon address table */
/* For all these resources, the version is defined as follows:
  0 = reserved
  2 = rio2
  3 = rio3
 */
CesResourceID ppcmon_table_vme[33] = {
  CESRT_SIO_1, /* serial */
  CESRT_LOCAL_CSR, /* loc_reg */
  CESRT_SIC,  /* int_ctl */
  (CesResourceID)0, /* iack */

  CESRT_SRAM, /* sram */
  (CesResourceID)0, /* nvram */
  CESRT_NOR_FLASH, /* fprom */
  (CesResourceID)0, /* sflash */

  (CesResourceID)0, /* timer */
  (CesResourceID)0, /* display */
  (CesResourceID)0, /* rtc */
  CESRT_XPC_CSR, /* csr */

  CESRT_VME_CSR, /* vme_csr */
  CESRT_DMA, /* vme_bma */
  CESRT_VME_A, /* vme_win0 */
  CESRT_VME_WIN1, /* vme_win1 */

  CESRT_LPCI_CSR, /* pci1_csr */
  CESRT_LPCI_IO,  /* pci1_io */
  CESRT_LPCI_MEM,  /* pci1_mem */
  CESRT_LPCI_CONF, /* pci1_conf */

  CESRT_XPCI_CSR, /* pci2_csr */
  CESRT_XPCI_IO,  /* pci2_io */
  CESRT_XPCI_MEM,  /* pci2_mem */
  CESRT_XPCI_CONF, /* pci2_conf */

  (CesResourceID)0, /* fifo_port */
  (CesResourceID)0, /* fifo_csr */
  (CesResourceID)0, /* fifo_mem */
  (CesResourceID)0, /* fifo_flag */

  CESRT_EPCI_CSR, /* pci3_csr */
  CESRT_EPCI_IO,  /* pci3_io */
  CESRT_EPCI_MEM,  /* pci3_mem */
  CESRT_EPCI_CONF, /* pci3_conf */

  CESRT_INVALID_RESOURCE /* End of table */
};
CesResourceID ppcmon_table_cpci[33] = {
  CESRT_SIO_1, /* serial */
  CESRT_LOCAL_CSR, /* loc_reg */
  CESRT_SIC, /* int_ctl */
  (CesResourceID)0, /* iack */

  CESRT_SRAM, /* sram */
  (CesResourceID)0, /* nvram */
  CESRT_NOR_FLASH, /* fprom */
  (CesResourceID)0, /* sflash */

  (CesResourceID)0, /* timer */
  (CesResourceID)0, /* display */
  (CesResourceID)0, /* rtc */
  CESRT_XPC_CSR, /* csr */

  CESRT_CPCI_CSR, /* pci1_csr */
  (CesResourceID)0,  /* pci1_io */
  (CesResourceID)0,  /* pci1_mem */
  (CesResourceID)0, /* pci1_conf */

  CESRT_LPCI_CSR, /* pci1_csr */
  CESRT_LPCI_IO,  /* pci1_io */
  CESRT_LPCI_MEM,  /* pci1_mem */
  CESRT_LPCI_CONF, /* pci1_conf */

  CESRT_XPCI_CSR, /* pci2_csr */
  CESRT_XPCI_IO,  /* pci2_io */
  CESRT_XPCI_MEM,  /* pci2_mem */
  CESRT_XPCI_CONF, /* pci2_conf */

  (CesResourceID)0, /* fifo_port */
  (CesResourceID)0, /* fifo_csr */
  (CesResourceID)0, /* fifo_mem */
  (CesResourceID)0, /* fifo_flag */

  CESRT_EPCI_CSR, /* pci3_csr */
  CESRT_EPCI_IO,  /* pci3_io */
  CESRT_EPCI_MEM,  /* pci3_mem */
  CESRT_EPCI_CONF, /* pci3_conf */

  CESRT_INVALID_RESOURCE /* End of table */
};

/* Set up one resource according to a table element */
void cesRT_Init_FromEntry(CesResourceTableEntry_t *resource, 
                                    CesResourceTableEntry_t *entry){

  cesRT_SetAddress(resource, entry->address);
  cesRT_SetSize(resource, entry->size);
  cesRT_SetVersion(resource, entry->version);
  cesRT_SetExtra(resource, entry->extra);
  cesRT_SetFlags(resource, entry->flags);

  return;
}

#if 0
/* This function is not considered to be needed */
/* Set up one resource according to the defaults table */
/* Needs the resource ID also */
CesError cesRT_Init_FromDefault(CesResourceTableEntry_t *resource,
                                      CesResourceID id){
  int i;

  i=0;
  /* Loop through defaults table to find matching line */
  while(defaultResourceDescriptors[i].id != CESRT_INVALID_RESOURCE) {
    if(defaultResourceDescriptors[i].id == id){
      cesRT_Init_FromEntry(resource,&defaultResourceDescriptors[i].entry);
      return CES_OK;
    }
    ++i;
    if( i >= CESRT_TABLE_SIZE) break;
  }
  return CES_ENOTFOUND;
}
#endif

/* Set up all resources  according to a defaults table */
/* this function takes a pointer to the defaults table,
   therefore it can be used with a different table */
void cesRT_Init_FromDescriptor(CesResourceTable_t *restable, 
                              CesResourceDescriptor_t *descriptor){
  int i;
  CesResourceTableEntry_t * resource;

  for (i = 0; i < CESRT_TABLE_SIZE; i++) {
    if (descriptor[i].id == CESRT_INVALID_RESOURCE)
      break;
    resource = (CesResourceTableEntry_t *) cesRT_GetResource(restable, descriptor[i].id);
    cesRT_Init_FromEntry(resource, &descriptor[i].entry);
  }
}

void cesRT_Init_EmptyTable(CesResourceTable_t *restable){
  int i;
  CesResourceTableEntry_t * resource;

  resource = (CesResourceTableEntry_t *) cesRT_GetResource(restable, CESRT_RESOURCE_TABLE);
  cesRT_SetAddress(resource, 0);
  cesRT_SetSize(resource, CESRT_LAST_RESOURCE * sizeof(CesResourceTableEntry_t));
  cesRT_SetVersion(resource, 0);
  cesRT_SetExtra(resource, CESRT_MAGIC);
  cesRT_SetFlags(resource, 0x00ff);
  for (i = 1; i < CESRT_LAST_RESOURCE; i++) {
    resource = (CesResourceTableEntry_t *) cesRT_GetResource(restable, (CesResourceID)i);
    cesRT_SetAddress(resource, 0xffffffff);
    cesRT_SetSize(resource, 0);
    cesRT_SetVersion(resource, 0);
    cesRT_SetExtra(resource, 0);
    cesRT_SetFlags(resource, 0);
  }
}

void cesRT_Init_FromPPCMon(CesResourceTable_t *restable, CesResourceID *ppcmon_table, void *fprom) {
  CesResourceTableEntry_t *resource;
  CesUint32 * address;
  int i;

  /* Physical address of the PPCMon table */
  address = fprom + 0x40;
  i=0;

/*  DBG("Updating table at 0x%08x from PPCMon table at 0x%08x\n",
   (CesUint32) restable, (CesUint32) address); */

  /* Loop on table */
  while(ppcmon_table[i] != CESRT_INVALID_RESOURCE) {
    if (ppcmon_table[i] != 0) {
      resource = (CesResourceTableEntry_t *) cesRT_GetResource(restable, ppcmon_table[i]);
/*      DBG(" Updating  0x%08x from 0x%08x value 0x%08x\n",
         (CesUint32) resource, (CesUint32) address, (CesUint32) *address); */
      if (*address != 0
       /*&& cesRT_GetFlags(resource) == 0
       && cesRT_GetFlags(resource) == CESRT_NONE*/)
        cesRT_SetAddress(resource, *address);
    }
    ++i;
    ++address;
    if(address >= (CesUint32 *)0xfff00100) break;
  }

  return;
}

CesUint8 cesRT_GetCpuFlag(void) {
  /* Static variable to store CPU flag */
  static int myCpuFlag = CESRT_NONE;

#ifdef VTHREADS
  myCpuFlag = CESRT_CPUA | CESRT_CPUB;
#else
  if(myCpuFlag == CESRT_NONE){
    if (cesHwIsRio4() && cesHwRio4IsBipro()) {
      switch(cesRT_GetCpuNumber()) {
        case 0: myCpuFlag = CESRT_CPUA;  break;
        case 1: myCpuFlag = CESRT_CPUB;  break;
        case 2: myCpuFlag = CESRT_CPUA1; break;
        case 3: myCpuFlag = CESRT_CPUB1; break;
      }
    } else {
      myCpuFlag = CESRT_CPUA | CESRT_CPUB;
    }
  }
#endif

  return myCpuFlag;
}

CesResourceTableEntry_t const * cesRT_GetPciSlot(CesResourceTable_t const *restable, CesUint32 bus_nr, CesUint32 slot_nr) {
  int res_id = 0;

  switch(bus_nr) {
    case 0: /* LPCI */
      if (slot_nr >= 2 && slot_nr <= 5)
        res_id = CESRT_LPCI_PMC1 + slot_nr - 2;
      if (slot_nr == 1) {/* LPCI eth */
        const CesResourceTableEntry_t * res;
        res = cesRT_GetResource(restable, CESRT_EPCI_ETH1);
        if (cesRT_GetVersion(res) == 0)
          return res;
      }
      break;
    case 1: /* XPCI */
      if (slot_nr <= 7)
        res_id = CESRT_XPCI_SLOT7 + slot_nr;
      break;
    case 2: /* CPCI */
      if (slot_nr <= 21)
        res_id = CESRT_CPCI_SLOT0 + slot_nr;
      break;
    case 3: /* EPCI */
      switch(slot_nr) {
        case 1: res_id = CESRT_EPCI_ETH1; break;
        case 2: res_id = CESRT_EPCI_ETH2; break;
      }
      break;
  }
  if (res_id == 0)
    return NULL;
  return cesRT_GetResource(restable, (CesResourceID)res_id);
}


static void cesRT_CopyFromNvram(void *addr, CesResourceDescriptor_t *table) {
  volatile char *src = addr;
  char *dst = (char*)table;
  int i;

  for (i = 0; i < sizeof(CesResourceDescriptor_t); i++) {
    dst[i] = src[i*4];
  }
}

static int cesRT_Init_FromNvram(CesResourceTable_t *restable, void *addr) {
  CesResourceDescriptor_t descriptor;
  CesResourceTableEntry_t *resource;
  int i;
  int size;

  cesRT_CopyFromNvram(addr + 0x2000, &descriptor);
  if (descriptor.id != 0)
    return 1;
  if (descriptor.entry.extra != CESRT_MAGIC)
    return 1;
  size = descriptor.entry.size / sizeof(CesResourceDescriptor_t);
  if (size * sizeof(CesResourceDescriptor_t) != descriptor.entry.size) {
    return 1;
  }
  for (i = 1; i < size; i++) {
    cesRT_CopyFromNvram(addr + 0x2000 + i * sizeof(CesResourceDescriptor_t), &descriptor);
    if (descriptor.id == CESRT_INVALID_RESOURCE)
      return 0;
    if (descriptor.id == CESRT_LAST_RESOURCE)
      continue;
    resource = (CesResourceTableEntry_t *) cesRT_GetResource(restable, descriptor.id);
    cesRT_Init_FromEntry(resource, &descriptor.entry);
  }
  return 0;
}

static int cesRT_Init_FromFram(CesResourceTable_t *restable, void *addr) {
  CesResourceDescriptor_t *descriptor;
  CesResourceTableEntry_t *resource;
  int i;
  int size;

  descriptor = addr + 0x800;
  if (descriptor->id != 0)
    return 1;
  if (descriptor->entry.extra != CESRT_MAGIC)
    return 1;
  size = descriptor->entry.size / sizeof(CesResourceDescriptor_t);
  if (size * sizeof(CesResourceDescriptor_t) != descriptor->entry.size) {
    return 1;
  }
  for (i = 1; i < size; i++) {
    descriptor = addr + 0x800 + i * sizeof(CesResourceDescriptor_t);
    if (descriptor->id == CESRT_INVALID_RESOURCE)
      return 0;
    if (descriptor->id == CESRT_LAST_RESOURCE)
      continue;
    resource = (CesResourceTableEntry_t *) cesRT_GetResource(restable, descriptor->id);
    cesRT_Init_FromEntry(resource, &descriptor->entry);
  }
  return 0;
}

void cesRT_Dump(CesResourceTable_t * const table, int(*print)(const char *fmt, ...)) {
/*	int count = 256; */
	int count = table->entries[0].size / sizeof(CesResourceTableEntry_t);

	int i;
	int j = 0;

	print("cesRT_GetTable() = {\n");
	for (i = 0;  i < count; i++) {
		CesResourceTableEntry_t *entry = &table->entries[i];
		if ((j % 40) == 0)
			print("/* id       address        size     ver   flags       extra  */\n");
		if ((CesUint32)entry->address != 0xffffffff) {
			print("{ %3d, { 0x%08x, 0x%08x, 0x%04x, 0x%04x, 0x%08x }},\n", i, (CesUint32)entry->address, entry->size, entry->version, entry->flags, entry->extra);
			j++;
		}
	}
	print("};\n");
}

CesError cesRT_Init(CesResourceTable_t *restable, void*(*phys_to_virt)(CesUint32 phys), CesUint32 offset) {
  void *fprom;
  int board_type;
  unsigned long nvram_phys;
  void *nvram;

  fprom = PHYS_TO_VIRT(0xfff00000);
  DBG_S("fprom is at "); DBG_H((int)fprom); DBG_C('\n');
  if (fprom == NULL)
     return CES_EIO;

  if (FPROM_PPCMON_VERSION(fprom) < 0x342e3630) {
    /* need PPC_Mon 4.60 or later! */
    return CES_EMISMATCH;
  }

  cesRT_Init_EmptyTable(restable);

  board_type = FPROM_BOARD_TYPE(fprom);
  DBG_S("board_type = "); DBG_H(board_type); DBG_C('\n');

  nvram_phys = FPROM_NVRAM_ADDR(fprom);
  nvram = PHYS_TO_VIRT(nvram_phys);
  DBG_S("nvram = "); DBG_H((int)nvram); DBG_C('\n');

  { /* Set board type */
    int btype;
    int i;
    CesResourceTableEntry_t *resource;
    resource = (CesResourceTableEntry_t *) cesRT_GetResource(restable, CESRT_BOARD_BASE);
    btype = 0;
    for (i = 0; i < 4; i++) {
      int charvalue = (board_type >> (i * 8)) & 0xff;
      if (charvalue >= '0' && charvalue <= '9') {
        btype |= (charvalue - '0') << (i * 4);
      } else if (charvalue >= 'A' && charvalue <= 'F') {
        btype |= (charvalue - 'A' + 10) << (i * 4);
      } else if (charvalue >= 'a' && charvalue <= 'f') {
        btype |= (charvalue - 'a' + 10) << (i * 4);
      } else {
        btype = board_type;
        break;
      }
    }
    cesRT_SetAddress(resource, btype);
    cesRT_SetSize(resource,    0); /* Memory size? */
    cesRT_SetVersion(resource, 0);
    cesRT_SetFlags(resource,   0);
    /* Serial Number */
    cesRT_SetExtra(resource, ((int)*((unsigned char*)nvram + 0x4038) << 8) | *((unsigned char*)nvram + 0x403c));
  }

  nvram = NULL;
  if (nvram != NULL) {
    if (BTYPE_IS_RIO4(board_type) ||
        BTYPE_IS_RIOS(board_type) ||
        BTYPE_IS_RIOC4070STYLE(board_type)) {
      if (!cesRT_Init_FromNvram(restable, nvram)) {
        DBG_S("NVRAM OK\n");
        return CES_OK;
      }
    } else {
      if (!cesRT_Init_FromFram(restable, nvram)) {
        DBG_S("FRAM OK\n");
        return CES_OK;
      }
    }
  }

  DBG_S("Falling back to hardware detection!\n");

  /* When we reach this point, we did not find any table in hardware.
     We need to detect things ourselves.
   */

  if (BTYPE_IS_RIO3(board_type)) {
    cesRT_Init_FromDescriptor(restable, PTR_OFFSET(rio3Resources));
    /* FIXME: The PPC_Mon entry for the 2nd serial line is wrong on RIO4!
       It says 0xf900b800 instead of 0xf900c800 */
    DBG_S("loading rio3Resources\n");
    if (BTYPE_IS_RIO3_VME(board_type)) {
      DBG_S("loading rio3VmeResources\n");
      cesRT_Init_FromDescriptor(restable, PTR_OFFSET(rio3VmeResources));
      cesRT_Init_FromPPCMon(restable, PTR_OFFSET(ppcmon_table_vme), fprom);
      if (NVRAM_BOARD_NAME(nvram, 4) == 'W') {
        DBG_S("loading rio3NoIrqRotationResources\n");
        cesRT_Init_FromDescriptor(restable, PTR_OFFSET(rio3NoIrqRotationResources));
      }
    } else {
      DBG_S("loading rio3CpciResources\n");
      cesRT_Init_FromDescriptor(restable, PTR_OFFSET(rio3CpciResources));
      cesRT_Init_FromDescriptor(restable, PTR_OFFSET(rio3NoIrqRotationResources));
      cesRT_Init_FromPPCMon(restable, PTR_OFFSET(ppcmon_table_cpci), fprom);
    }
    if (restable->entries[CESRT_XPCI_CSR].address == CESRT_NONE) {
      DBG_S("loading rio3NoIrqRotationResources\n");
      cesRT_Init_FromDescriptor(restable, PTR_OFFSET(rio3NoIrqRotationResources));
    }
  } else if(BTYPE_IS_RIO4(board_type)) {
    /* We need this as long as there is no real table in hardware. */
    DBG_S("loading rio4Resources\n");
    cesRT_Init_FromDescriptor(restable, PTR_OFFSET(rio4Resources));
  } else if (BTYPE_IS_RIOS(board_type)) {
    DBG_S("loading rioSResources\n");
    cesRT_Init_FromDescriptor(restable, PTR_OFFSET(rioSResources));
  } else if ( BTYPE_IS_RIOC4070STYLE(board_type) ) {
    DBG_S("loading rioc4070styleResources\n");
    cesRT_Init_FromDescriptor(restable, PTR_OFFSET(rio3Resources));
    cesRT_Init_FromDescriptor(restable, PTR_OFFSET(rioc4070styleResources));
    if ( board_type == CES4070 ) {
      DBG_S("loading rioc4070Resources\n");
      cesRT_Init_FromDescriptor(restable, PTR_OFFSET(rioc4070Resources));
    }
  } else if ( BTYPE_IS_CLASSICMFCC(board_type) ) {
    DBG_S("loading classicMfccResources\n");
    cesRT_Init_FromDescriptor(restable, PTR_OFFSET(classicMfccResources));
  } else if ( BTYPE_IS_8463STYLEMFCC(board_type) ) {
    cesRT_Init_FromDescriptor(restable, PTR_OFFSET(mfcc8463styleResources));
  }
  return CES_OK;
}

CesError cesRT_SaveModifiedTable(const CesResourceTable_t *table) {
  /* write modified table to sram */
  /* If there is a table in nvram, write the modified table to nvram */
  /* FIXME: Is it desired to create a table in nvram if there does not yet exist one? */
  /* Basically we need to overwrite the table which we were reading because:
     - if there is no table in nvram, the second CPU will see the sram table only
     - if there is one in nvram, the second CPU will see the nvram table only
     There are 2 possible solutions to know where we have read the table:
     1. remember where we have read the table from in an entry of the resource table
     2. do the searching again.
   */
  /* FIXME: How should this function get the virtual address of nvram/sram?
            Possible solutions:
            - by calling cesPhysToVirt()
            - by calling cesIoPhysToVirt()   (seems more consequent to separate I/O from memory phys<->virt conversion)
            - getting the nvram address as an argument
   */
  return CES_ENOTIMPL;
}
