/*
 * Copyright (C) Rupert Eibauer <ces@rupert-eibauer.de>
 *
 *
 * Rewritten from rio_pci.c(2.2) apus_pci.c and pmac_pci.c
 *
 *
 */

#include <linux/autoconf.h>

#include <linux/kernel.h>
#include <linux/pci.h>
#include <linux/init.h>
#include <linux/delay.h>
#include <linux/string.h>
#include <linux/version.h>
#include <linux/ioport.h>

#include <asm/io.h>
#include <asm/machdep.h>
#include <asm/pci.h>
#include <asm/pci-bridge.h>
#include <asm/sections.h>

#include "rio.h"

#include <ces/cesResourceTable.h>

/* FIXME: Please check if these are defined in rio.h */
#define RIO_WRITE_SGO_PAGES
//#define USE_PREF_MEMORY
//#define PCI_SLAVE_IS_UPDATED
//#define RIO_IMPLEMENT_IOMMU

#define SGO_TYPE_D64  0x02

#define SGO_TYPE_STD  0x00
#define SGO_TYPE_RRMW 0x04
#define SGO_TYPE_RLWI 0x08

#define SGO_TYPE_MEM  0x00
#define SGO_TYPE_IO   0x40
#define SGO_TYPE_CFG  0x60

#define IO_BUS  0x00000000
#define MEM_BUS 0xc0000000
#define CFG_BUS 0x80000000


/* FIXME: These two are ideas which are not yet implemented */
/* Could be smaller:
  - an i386 has 65536 bytes
  - an MFCC has 4MB! (4 of them can be on an MFCC!)
    Then, 16MB is required. Note that these are __REALLY__ old
    mfcc, which should not be used anymore.
 */
#define IO_MAP_SIZE   RIO_SGO_PAGE_SIZE

/* Could be smaller:
	 64k =  1 bus (primary only)
	128k =  2 busses
	256k =  4 busses
	512k =  8 busses
	  1M = 16 busses
	It could also be mapped on-demand per-bus inside of the
	pci_ops.
 */
#define CFG_MAP_SIZE  RIO_SGO_PAGE_SIZE

/*
 * RIO PCI setup and configuration.
 */
#if 0
This is only for looking at.
it has moved into <asm/rio.h>

typedef struct
{
	unsigned int pci[0x20];	 /* 0x00 */
	unsigned int ctl;	 /* 0x80 */
	unsigned int sts;	 /* 0x84 */
	unsigned int sg_addr;	 /* 0x88 */
	unsigned int sg_data;	 /* 0x8c */
	unsigned int rsv[4];
	unsigned int int_mask;	 /* 0xa0 */
	unsigned int int_pend;	 /* 0xa4 */
	unsigned int rsv1;
	unsigned int int_gen;	 /* 0xac */
	unsigned int rsv2[4];
	unsigned int loc_sem[8]; /* 0xc0 */
	unsigned int loc_ram[8]; /* 0xe0 */
} rio_pci_csr_t;
#endif

#define MB (1024*1024)
/* FIXME: Make this configuration more dynamic */
rio_pci_cfg_t rio_pci_cfg[RIO_PCI_BUSSES] = {
	{
		.name = "LPCI Bridge",
		.enabled   = 1,
		.sys_slot  = 1,
		.csr       = (rio_pci_csr_t*)0x000, /* Offset. The virt addr will be added later. */
		.max_dev   = 3, /* 5 */
		.sgo_base  = 0xc0000000,
		.sgo_count = 0x38,
		.sgi_psize = 16*MB,
		.sgi_end   = 0,
		.sgo_top   = -1,
		.sgi_count = 0x100
	}, {
		.name = "XPCI Bridge",
		.enabled   = 0,
		.sys_slot  = 0,
		.csr       = (rio_pci_csr_t*)0x100, /* Offset. The virt addr will be added later. */
		.max_dev   = 3,
		.sgo_base  = 0x80000000,
		.sgo_count = 0x40,
		.sgi_psize = 1*MB,
		.sgi_end   = 0,
		.sgo_top   = -1,
		.sgi_count = 0x100
	}, {
		.name = "CPCI Bridge",
		.enabled   = 0,
		.sys_slot  = 0,
		.csr       = (rio_pci_csr_t*)0x200, /* Offset. The virt addr will be added later. */
		.max_dev   = 21,
		.sgo_base  = 0x40000000,
		.sgo_count = 0x40,
		.sgi_psize = 1*MB,
		.sgi_end   = 0,
		.sgo_top   = -1,
		.sgi_count = 0x100
	}, {
		.name = "EPCI Bridge",
		.enabled   = 0,
		.sys_slot  = 1,
		.csr       = (rio_pci_csr_t*)0x400, /* Offset. The virt addr will be added later. */
		.max_dev   = 2,
		.sgo_base  = 0xa0000000,
		.sgo_count = 0x8,
		.sgi_psize = 16*MB,
		.sgi_end   = 0,
		.sgo_top   = -1,
		.sgi_count = 0x100
	}
};

EXPORT_SYMBOL(rio_pci_cfg);

static rio_pci_cfg_t * lpci_cfg = &rio_pci_cfg[0];
static rio_pci_cfg_t * xpci_cfg = &rio_pci_cfg[1];
static rio_pci_cfg_t * cpci_cfg = &rio_pci_cfg[2];
static rio_pci_cfg_t * epci_cfg = &rio_pci_cfg[3];

/* At some point, this should be integrated into rio_pci_cfg_t */
struct pci_slot_info_t {
	char ints[4];
	char visible;
};
static struct pci_slot_info_t lpci_slot_info[9];
static struct pci_slot_info_t xpci_slot_info[8];
static struct pci_slot_info_t cpci_slot_info[21];
static struct pci_slot_info_t epci_slot_info[5];
static struct pci_slot_info_t *pci_slot_info[4] = {lpci_slot_info, xpci_slot_info, cpci_slot_info, epci_slot_info};
static int pci_slot_count[4] = { 9, 8, 21, 5 };

/*
 *   The following code swizzles for exactly one bridge.  The routine
 *   common_swizzle below handles multiple bridges.  But there are a
 *   some boards that don't follow the PCI spec's suggestion so we
 *   break this piece out separately.
 */
static inline unsigned char bridge_swizzle(unsigned char pin,
                unsigned char idsel)
{
	return (((pin-1) + idsel) % 4) + 1;
}

/*
 * Re-route the interrupt for the devices.
 */
static inline u8
peb6418_swizzle(unsigned char pin, unsigned char idsel)
{
	if (idsel & 5) {
		/* idsel 1 or 4: Top (PMC2): do nothing */
		/* pin = pin; */
	} else  {
		/* idsel 2 or 8: Bottom (PMC1): shift by 2 */
		pin  = ((pin + 1) % 4) + 1;
	}
	return pin;
}

u8 __devinit
rio_swizzle(struct pci_dev *dev, u8 *pinp)
{
	struct pci_controller *hose = dev->sysdata;
	rio_pci_cfg_t *cfg = hose->arch_data;

	if (dev->bus != hose->bus) {
		u8 pin = *pinp;
		do {
			/* We assume that it's a PEB */
			/* FIXME: better check for PEB? */
			if (cfg == xpci_cfg && dev->bus->parent && !dev->bus->parent->parent) {
				pin = peb6418_swizzle(pin, PCI_SLOT(dev->devfn));
			} else {
				pin = bridge_swizzle(pin, PCI_SLOT(dev->devfn));
			}
			/* Move up the chain of bridges. */
			dev = dev->bus->self;
		} while (dev->bus->self);
		*pinp = pin;

		/* The slot is the idsel of the last bridge. */
	}
	return PCI_SLOT(dev->devfn);
}

int __devinit
rio_map_irq(struct pci_dev *dev, u8 slot, u8 pin)
{
	struct pci_controller *hose = dev->sysdata;
	rio_pci_cfg_t *cfg = hose->arch_data;
	int retval = 0;
	int id = cfg - rio_pci_cfg;

	retval = pci_slot_info[id][slot].ints[(pin - 1) % 4];
#ifdef CONFIG_RIO_XPCI_INTS
#error CONFIG_RIO_XPCI_INTS NOT IMPLEMENTED
	if (cfg == xpci_cfg && dev->bus->number == hose->first_busno) {
		if (slot == 3)
			retval = RIO_ACT_XPCI3_INTA + pin - 1;
		else
			retval = RIO_ACT_XPCI_INTA + pin - 1;
	}
#endif

	if (retval == 0) {
		printk(KERN_WARNING "%s: could not map irq of device %s\n", __FUNCTION__, pci_name(dev));
	}

	return retval;
}


#if 0
/* FIXME: We should trust  on PPC_Mon to get this
   device right. */
static void fixup_8442(struct pci_dev *dev) {
	struct pci_controller *hose = dev->sysdata;
	if (dev->resource[1].start == 0 ||
		dev->resource[1].end == -1 ||
		dev->resource[1].flags != 0x101) {
		char rev;
		unsigned long size;
		pci_read_config_byte(dev, 0x08, &rev);
		if (rev & 0x80)
			size=256;
		else
			size=4*1024*1024;
		dev->resource[1].flags = 0x101;
		dev->resource[1].end = dev->resource[1].start + size - 1;
		dev->resource[1].start += hose->io_resource.start;
		dev->resource[1].end += hose->io_resource.start;
		pcibios_update_resource(dev, NULL, &dev->resource[1], 1);
		printk("RIO_PCI: Fixed up IO size of MFCC %x rev %x!\n", dev->devfn, rev);
	}
}
DECLARE_PCI_FIXUP_HEADER(0x10d9, 0x8442, fixup_8442);
#endif

static void fixup_cmd_646_pmc(struct pci_dev *dev) {
	char tmp;
	printk("RIO_PCI: Disable 2nd channel of CMD646 in slot %s\n", pci_name(dev));
	pci_read_config_byte(dev, 0x51, &tmp);
	tmp &= ~0x80;
	pci_write_config_byte(dev, 0x51, tmp);
}
DECLARE_PCI_FIXUP_HEADER(PCI_VENDOR_ID_CMD, PCI_DEVICE_ID_CMD_646, fixup_cmd_646_pmc);

static void fixup_peb(struct pci_dev *dev) {
	struct pci_controller *hose = dev->sysdata;
	rio_pci_cfg_t *cfg = hose->arch_data;
	if (cfg == xpci_cfg) {
		unsigned long io_start;
		unsigned long io_end;
		unsigned long mem_start;
		unsigned long mem_end;
		u32 l;

//		printk("Fixup PEB PCI Bridge devfn=%x\n", dev->devfn);

		pci_read_config_dword(dev, PCI_IO_BASE, &l);
		l &= 0xffff0000;

		if (dev->devfn==0x10) { // 1st PEB    (maybe we can do that in fixup_bus() ?
			io_start = 0x01800000;
			io_end =   0x01ffffff;
			mem_start = 0x9f000000;
			mem_end =   0xbdffffff;
		} else if (dev->devfn==0x18) { // 2nd PEB
			io_start = 0x01000000;
			io_end =   0x017fffff;
			mem_start = 0x80000000;
			mem_end =   0x9effffff;
			// Adjust allocation base for 2nd PEB (dirty Hack!)
			cfg->io_top = 0x01800000;
			cfg->mem_top = 0x9f000000;
		} else {
			printk("%s: Invalid devfn %x for PEB!\n", __FUNCTION__, dev->devfn);
			return;
		}
		l |= PCI_IO_RANGE_TYPE_32 | ((io_start >> 8) & 0x00f0) | (io_end & 0xf000);
//		printk("PCI_IO_BASE Register = %x\n", l);
		pci_write_config_dword(dev, PCI_IO_BASE, l);
		pci_write_config_dword(dev, PCI_IO_BASE_UPPER16,
			((io_start >> 16) & 0xffff) |
			(io_end & 0xffff0000)        );
#ifdef USE_PREF_MEMORY
		pci_write_config_dword(dev, PCI_PREF_MEMORY_BASE,
			((mem_start >> 16) & 0xfff0) |
			(mem_end & 0xfff00000)        );
		pci_write_config_dword(dev, PCI_MEMORY_BASE, 0xfff0);
#else
		pci_write_config_dword(dev, PCI_MEMORY_BASE,
			((mem_start >> 16) & 0xfff0) |
			(mem_end & 0xfff00000)        );
		pci_write_config_dword(dev, PCI_PREF_MEMORY_BASE, 0xfff0);
#endif
		pci_write_config_word(dev, PCI_COMMAND, PCI_COMMAND_IO | PCI_COMMAND_MEMORY | PCI_COMMAND_MASTER);
		pci_write_config_byte(dev, PCI_CACHE_LINE_SIZE, 8);
		pci_write_config_byte(dev, PCI_LATENCY_TIMER, 0xff);
	}
}
DECLARE_PCI_FIXUP_HEADER(0x1011, 0x0026, fixup_peb);


static void rio_pcibios_fixup(void);

static void
__init rio_setup_pci_ptrs(void)
{
#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,9)
	int i;
#error This code is obsolete!
	for (i=0; pcibios_fixups[i].pass!=0; i++) {
		if (pcibios_fixups[i].vendor == (u16)PCI_ANY_ID &&
		    pcibios_fixups[i].device == (u16)PCI_ANY_ID) {
			pcibios_fixups[i].hook = rio_fixup_resources;
//			printf("Setting up pcibios_fixup[%d].hook\n", i);
			return;
		}
	}
#endif
	ppc_md.pci_swizzle = rio_swizzle;
	ppc_md.pci_map_irq = rio_map_irq;
	ppc_md.pcibios_fixup = rio_pcibios_fixup;
}

int __init rio_pci_slave_init(rio_pci_cfg_t* cfg)
{
	int i, size, count, page, space;
	unsigned int addr;
	int addr_shift;
	int wbsel = 0;
	CesResourceTable_t const *restable;
	CesResourceTableEntry_t const * entry;

	restable = cesRT_GetTable();

	if (cfg == lpci_cfg) { /* LPCI */
		size = rio_board->memory_size;
		wbsel = 0x80000000;
	} else if (cfg == epci_cfg) { /* EPCI */
		u32 csr;

		/* Set the slave prefetch size to maximum! */
		csr = readl(&cfg->csr->ctl);
		csr |= 0x00000060;
		writel(csr, &cfg->csr->ctl);

		size = rio_board->memory_size;

	} else { /* XPCI or CPCI */
		unsigned int csr;
		if (cfg->sys_slot) {
			int divisor = 0x4000000;
			printk("%s: PCI Slave Setup for System Slot\n", __FUNCTION__);
			/* Set slave size = memory size */
			size = rio_board->memory_size;
#ifndef PCI_SLAVE_IS_UPDATED
			if (size > (256*MB)) {
				size = 256*MB;
				printk("WARNING: %cPCI slave size limited to 256MB\n",
				       cfg->name[0]);
			}
#endif
			/* Set slave windown size */
			csr = readl(&cfg->csr->ctl);
			csr &= ~0x1c;
			while ((size / divisor)) {
				divisor *= 2;
				csr += 0x4;
			}
			writel(csr, &cfg->csr->ctl);
		} else {
			printk("RIO_PCI: PCI Slave Setup for Slave Slot\n");
			/* Get slave window size */
			csr = (readl(&cfg->csr->ctl) >> 2) & 0x7;
			size = (32*MB) << csr;
#ifndef PCI_SLAVE_IS_UPDATED
			if (size > (256*MB)) {
				size = 256*MB;
				printk("WARNING: %cPCI slave size limited to 256MB\n",
				       cfg->name[0]);
			}
#endif
		}
#ifdef PCI_SLAVE_IS_UPDATED
		/* Scale page size if > 256 MB */
		if (size > (512*MB))
			cfg->sgi_psize *= 4;
		if (size > (256*MB))
			cfg->sgi_psize *= 4;
#endif
		printk("RIO_PCI: Setting PCI mem base address to 0\n");
		/* Set PCI mem base address to 0 */
		writel(0x0, &cfg->csr->pci[4]);
		/* Enable PCI mem space and PCI master */
		writel(0x6, &cfg->csr->pci[1]);
	}

	count = size / cfg->sgi_psize;
	printk("rio_pci_slave_setup (%s) size=0x%08x psize=0x%08x count=%d\n", cfg->name, size, cfg->sgi_psize, count);
//	printk(KERN_ERR "RIO_PCI SGI: addr 0x%08x count 0x%x space 0x%x\n", addr, count, space);

	entry = cesRT_GetResource(restable, CESRT_LPCI_CSR);
	if (cesRT_GetVersion(entry) == 1) {
		page = pci_dram_offset / cfg->sgi_psize; /* Use the DRAM offset (which is also used by the mapping functions) */
		space = rio_board->cpu_nr ? 2 : 0;
		/* Snooping policy.
		   0x8 = Snooping with line FLUSH
		   0xc = Snooping with line KILL
		   The latter one may be more efficient, and I think
		   the Linux kernel does separate dma buffers well
		   enough, but for now we live with FLUSH to be sure.
		   Rupert
		*/
		space |= 0x80 | wbsel; /* snooping */
		addr_shift = 8;
	} else if (cesRT_GetVersion(entry) == 0) {
		page = 0;
		space = 0x2;
		addr_shift = 16;
	} else {
		printk(KERN_ERR "%s: Don't know how to program SGI\n", __func__);
		return count;
	}

	addr = 0;
	for (i = 0; i < count; i++) {
		out_le32(&cfg->csr->sg_addr, 0x10000 + page + i);
//		printk("ssram_write(0x%08x, 0x%08x -> 0x%08x)\n", 0x10000 + page + i, in_le32(&cfg->csr->sg_data), (addr >> addr_shift) | space);
		out_le32(&cfg->csr->sg_data, (addr >> addr_shift) | space);
		addr += cfg->sgi_psize;
	}
	return count;
}

#ifdef RIO_IMPLEMENT_IOMMU
void rio_pci_map_coherent(struct pci_device *dev, void *addr, int size) {
	struct pci_controller *hose = dev->sysdata;
	rio_pci_cfg_t *cfg = hose->arch_data;
	unsigned long dma_address;

	dma_address = xpc_pci_slave_map(cfg - rio_pci_cfg, virt_to_phys(xpc_address), size, XPC_PCI_SPACE_MEM, XPC_ADDRESS_ANY);
	return dma_address;
}

dma_addr_t pci_dma_map_single(struct pci_device *dev, void *ptr, size_t size,
	enum dma_data_direction direction)
{
	struct pci_controller *hose;
	rio_pci_cfg_t *cfg;

	if (!dev) {
		return -1;
	}
	hose = dev->sysdata
	cfg = hose->arch_data

	return xpc_pci_slave_map(cfg - rio_pci_cfg, virt_to_phys(ptr), size, XPC_PCI_SPACE_MEM, XPC_ADDRESS_ANY);
}
void pci_dma_unmap_single(struct pci_device *dev, dma_addr_t dma_addr, size_t size,
	enum dma_data_direction direction)
{
	struct pci_controller *hose;
	rio_pci_cfg_t *cfg;

	if (!dev) {
		return -1;
	}
	hose = dev->sysdata
	cfg = hose->arch_data

	xpc_pci_slave_unmap(cfg - rio_pci_cfg, dma_addr, size);
}
void rio_pci_slave_page_setup(rio_pci_cfg_t *cfg, int xpcaddr, int pciaddr, int space) {
	CesResourceTable_t *restable;
	CesResourceTableEntry_t const * entry;

	restable = cesRT_GetTable();

	entry = cesRT_GetResource(restable, CESRT_LPCI_CSR);
	if (cesRT_GetVersion(entry) == 1) {
		pciaddr += pci_dram_offset / cfg->sgi_psize; /* Use the DRAM offset (which is also used by the mapping functions) */
		space |= rio_board->cpu_nr ? 2 : 0;
		/* Snooping policy.
		   0x8 = Snooping with line FLUSH
		   0xc = Snooping with line KILL
		   The latter one may be more efficient, and I think
		   the Linux kernel does separate dma buffers well
		   enough, but for now we live with FLUSH to be sure.
		   Rupert
		*/
		space |= 0x80 | wbsel; /* snooping */
		addr_shift = 8;
	} else {
		space = 0x2;
		addr_shift = 16;
	}

	out_le32(&cfg->csr->sg_addr, 0x10000 + xpcaddr / cfg->sgi_psize);
	out_le32(&cfg->csr->sg_data, (pciaddr >> addr_shift)  | space); /* */
}
#endif /* RIO_IMPLEMENT_IOMMU */

#ifdef RIO_WRITE_SGO_PAGES
static unsigned int __init rio_pci_sgo_write(rio_pci_cfg_t* cfg, int page,
				      int count, int space,
				      unsigned int addr)
{
	int i = page * 4, j, data;
//	printk("rio_pci_sgo_write(\"%s\", page %d, count %d, space %x, addr %x)\n", cfg->name, page, count, space, addr);

	for (j = 0; j < count; j++, i+= 4, addr += RIO_SGO_PAGE_SIZE) {
		data = space | (addr >> 16);
		out_le32(&cfg->csr->sg_addr, i);
//		printk("sram[%d]: old: %x new: %x\n", i, in_le32(&cfg->csr->sg_data), data);
		out_le32(&cfg->csr->sg_data, data);
		out_le32(&cfg->csr->sg_addr, i + 1);
		out_le32(&cfg->csr->sg_data, 0);
		out_le32(&cfg->csr->sg_addr, i + 2);
		out_le32(&cfg->csr->sg_data, 0);
		out_le32(&cfg->csr->sg_addr, i + 3);
		out_le32(&cfg->csr->sg_data, 0);
	}
	return cfg->sgo_base + page * RIO_SGO_PAGE_SIZE;
}
#endif

/* Resource: */
/* 0 = MEM space (or VME) */
/* 1 = I/O space */
/* 2 = CFG space */

/* Only for PCI busses for now */
#define BT(b, t) ((b << 2) | t)
CesResourceTableEntry_t const *rio_get_bus_resource(int bus_nr, int type) {
	const CesResourceTable_t *lResourceTable = cesRT_GetTable();
	CesResourceTableEntry_t const * lRes;
	CesResourceID lResId;

	switch (BT(bus_nr, type)) {
		case BT(0, 0): lResId = CESRT_LPCI_MEM; break;
		case BT(0, 1): lResId = CESRT_LPCI_IO; break;
		case BT(0, 2): lResId = CESRT_LPCI_CONF; break;
		case BT(0, 3): lResId = CESRT_LPCI_CSR; break;

		case BT(1, 0): lResId = CESRT_XPCI_MEM; break;
		case BT(1, 1): lResId = CESRT_XPCI_IO; break;
		case BT(1, 2): lResId = CESRT_XPCI_CONF; break;
		case BT(1, 3): lResId = CESRT_XPCI_CSR; break;

		case BT(2, 0): lResId = CESRT_CPCI_MEM; break;
		case BT(2, 1): lResId = CESRT_CPCI_IO; break;
		case BT(2, 2): lResId = CESRT_CPCI_CONF; break;
		case BT(2, 3): lResId = CESRT_CPCI_CSR; break;

		case BT(3, 0): lResId = CESRT_EPCI_MEM; break;
		case BT(3, 1): lResId = CESRT_EPCI_IO; break;
		case BT(3, 2): lResId = CESRT_EPCI_CONF; break;
		case BT(3, 3): lResId = CESRT_EPCI_CSR; break;
		default: return NULL;
	}
	lRes = cesRT_GetResource(lResourceTable, lResId);
	if (cesRT_IsMy(lRes)) {
//		printk("rio_get_bus_resource(%d, %d): start=%08lx size=%08x version=%04x flags=%04x extra=%08x\n", bus_nr, type, lRes->address, lRes->size, lRes->version, lRes->flags, lRes->extra);
		return lRes;
	}
#if 0
	/* If that failed, try the secondary resources */
	switch (BT(bus_nr, type)) {
		case BT(0, 0): lResId = CESRT_LPCI_MEM2; break;
		case BT(0, 1): lResId = CESRT_LPCI_IO2; break;

		case BT(1, 0): lResId = CESRT_XPCI_MEM2; break;
		case BT(1, 1): lResId = CESRT_XPCI_IO2; break;

		case BT(2, 0): lResId = CESRT_CPCI_MEM2; break;
		case BT(2, 1): lResId = CESRT_CPCI_IO2; break;

		case BT(3, 0): lResId = CESRT_EPCI_MEM2; break;
		case BT(3, 1): lResId = CESRT_EPCI_IO2; break;
		default: return;
	}
	lRes = cesRT_GetResource(lResourceTable, lResId);
	if (cesRT_IsMy(lRes)) {
		*start = cesRT_GetAddress(lRes);
		*size  = cesRT_GetSize(lRes);
	}
#endif
//	printk("rio_get_bus_resource(%d, %d): no Resource found!\n", bus_nr, type);
	return NULL;
}
#undef BT

void rio_get_bus_window(int bus_nr, int type, int *start, int *size) {
	CesResourceTableEntry_t const *lRes;
	int flags;

	lRes = rio_get_bus_resource(bus_nr, type);

	*start = cesRT_GetAddress(lRes);
	*size = cesRT_GetSize(lRes);

	if (type >= 2)
		return;

	flags = cesRT_GetFlags(lRes);
	if (flags & 0x8000) {
		int lSize = 1;
		while (lSize < *size)
			lSize <<= 1;
		lSize >>= 8;
		lSize = (cesRT_GetExtra(lRes) & 0xff) * lSize;
		if (cesRT_GetCpuNumber() == 0) {
			*size = lSize;
		} else {
			*size -= lSize;
			*start += lSize;
		}
	}
}

static void __init pci_init_resource(struct resource *res, resource_size_t start,
			      resource_size_t end, int flags, char *name)
{
	res->start = start;
	res->end = end;
	res->flags = flags;
	res->name = name;
	res->parent = NULL;
	res->sibling = NULL;
	res->child = NULL;
}

#define rio_init_resource(res, name, start, len, flags) pci_init_resource(res, start, (start)+(len)-1, flags, name)

static int pci_slot_is_enabled(rio_pci_cfg_t *cfg, int devno) {
	int id = cfg - rio_pci_cfg;
	if (devno >= pci_slot_count[id])
		return 0;
	return pci_slot_info[id][devno].visible;
}

static void rio_init_pci_slot(struct pci_slot_info_t *info, CesResourceID id) {
	CesResourceTableEntry_t const * lRes;

	lRes = cesRT_GetResource(cesRT_GetTable(), id);
	info->ints[0] = (cesRT_GetAddress(lRes) >> 24) & 0xff;
	info->ints[1] = (cesRT_GetAddress(lRes) >> 16) & 0xff;
	info->ints[2] = (cesRT_GetAddress(lRes) >>  8) & 0xff;
	info->ints[3] = (cesRT_GetAddress(lRes) >>  0) & 0xff;
	info->visible = cesRT_IsMy(lRes);
}

static void rio_init_pci_slots(void) {
	rio_init_pci_slot(&lpci_slot_info[1], CESRT_LPCI_PMC0);
	rio_init_pci_slot(&lpci_slot_info[2], CESRT_LPCI_PMC1);
	rio_init_pci_slot(&lpci_slot_info[3], CESRT_LPCI_PMC2);
	rio_init_pci_slot(&lpci_slot_info[4], CESRT_LPCI_PMC3);
	rio_init_pci_slot(&lpci_slot_info[5], CESRT_LPCI_PMC4);
	rio_init_pci_slot(&lpci_slot_info[6], CESRT_LPCI_PMC5);
	rio_init_pci_slot(&lpci_slot_info[7], CESRT_LPCI_PMC6);
	rio_init_pci_slot(&lpci_slot_info[8], CESRT_LPCI_PMC7);

	rio_init_pci_slot(&xpci_slot_info[0], CESRT_XPCI_SLOT0);
	rio_init_pci_slot(&xpci_slot_info[1], CESRT_XPCI_SLOT1);
	rio_init_pci_slot(&xpci_slot_info[2], CESRT_XPCI_SLOT2);
	rio_init_pci_slot(&xpci_slot_info[3], CESRT_XPCI_SLOT3);
	rio_init_pci_slot(&xpci_slot_info[4], CESRT_XPCI_SLOT4);
	rio_init_pci_slot(&xpci_slot_info[5], CESRT_XPCI_SLOT5);
	rio_init_pci_slot(&xpci_slot_info[6], CESRT_XPCI_SLOT6);
	rio_init_pci_slot(&xpci_slot_info[7], CESRT_XPCI_SLOT7);

	rio_init_pci_slot(&cpci_slot_info[ 0], CESRT_CPCI_SLOT0);
	rio_init_pci_slot(&cpci_slot_info[ 1], CESRT_CPCI_SLOT1);
	rio_init_pci_slot(&cpci_slot_info[ 2], CESRT_CPCI_SLOT2);
	rio_init_pci_slot(&cpci_slot_info[ 3], CESRT_CPCI_SLOT3);
	rio_init_pci_slot(&cpci_slot_info[ 4], CESRT_CPCI_SLOT4);
	rio_init_pci_slot(&cpci_slot_info[ 5], CESRT_CPCI_SLOT5);
	rio_init_pci_slot(&cpci_slot_info[ 6], CESRT_CPCI_SLOT6);
	rio_init_pci_slot(&cpci_slot_info[ 7], CESRT_CPCI_SLOT7);
	rio_init_pci_slot(&cpci_slot_info[ 8], CESRT_CPCI_SLOT8);
	rio_init_pci_slot(&cpci_slot_info[ 9], CESRT_CPCI_SLOT9);
	rio_init_pci_slot(&cpci_slot_info[10], CESRT_CPCI_SLOT10);
	rio_init_pci_slot(&cpci_slot_info[11], CESRT_CPCI_SLOT11);
	rio_init_pci_slot(&cpci_slot_info[12], CESRT_CPCI_SLOT12);
	rio_init_pci_slot(&cpci_slot_info[13], CESRT_CPCI_SLOT13);
	rio_init_pci_slot(&cpci_slot_info[14], CESRT_CPCI_SLOT14);
	rio_init_pci_slot(&cpci_slot_info[15], CESRT_CPCI_SLOT15);
	rio_init_pci_slot(&cpci_slot_info[16], CESRT_CPCI_SLOT16);
	rio_init_pci_slot(&cpci_slot_info[17], CESRT_CPCI_SLOT17);
	rio_init_pci_slot(&cpci_slot_info[18], CESRT_CPCI_SLOT18);
	rio_init_pci_slot(&cpci_slot_info[19], CESRT_CPCI_SLOT19);
	rio_init_pci_slot(&cpci_slot_info[20], CESRT_CPCI_SLOT20);
	rio_init_pci_slot(&cpci_slot_info[21], CESRT_CPCI_SLOT21);

	rio_init_pci_slot(&epci_slot_info[1], CESRT_EPCI_ETH1);
	rio_init_pci_slot(&epci_slot_info[2], CESRT_EPCI_ETH2);
	rio_init_pci_slot(&epci_slot_info[3], CESRT_EPCI_ETH3);
	rio_init_pci_slot(&epci_slot_info[4], CESRT_EPCI_ETH4);
}

static struct pci_ops rio_pci_ops;

void __init rio_init_pci_bus(int i) {
#ifdef RIO_WRITE_SGO_PAGES
	unsigned long io_phys;
	unsigned long mem_phys;
#endif
	u32 start, page;
	int size, count;
	struct pci_controller *hose;
	rio_pci_cfg_t* cfg;

	cfg = &rio_pci_cfg[i];
	if (cfg == lpci_cfg) {
		/* Set the flushing mode for lpci (Required for FPCOM support) */
		/* FIXME: Should we do this for all busses? */
		int ctl;
		ctl = readl(&cfg->csr->ctl);
		/* Which one is correct? */
		//ctl |= 0x00000080;
		ctl &= ~0x00000080;
		writel(ctl, &cfg->csr->ctl);
	}

	if (!cfg->enabled)
		return;

	cfg->sgi_end = rio_pci_slave_init(cfg);

	if (!cfg->sys_slot)
		return;

	hose = pcibios_alloc_controller(NULL); /* The dev arg ends up in arch_data */
	if (hose == NULL) {
		printk(KERN_ERR "%s: Unable to allocate hose\n", __FUNCTION__);
		cfg->enabled = 0;
		return;
	}

	hose->ops = &rio_pci_ops;
	hose->arch_data = cfg;
	hose->first_busno = i << 4;


	cfg = &rio_pci_cfg[i];

	/*********** CFG Window ***********/
	/* FIXME: maybe only ioremap starting from our bus number + 16 busses? */
	rio_get_bus_window(i, 2, &start, &size);
	page = (start - cfg->sgo_base) / RIO_SGO_PAGE_SIZE;
	count = size / RIO_SGO_PAGE_SIZE;
#ifdef RIO_WRITE_SGO_PAGES
	rio_pci_sgo_write(cfg, page, count, SGO_TYPE_CFG, CFG_BUS);
#endif
	hose->cfg_addr = ioremap(start, size);
//	printk("%s: %s setting hose(%p)->cfg_addr to phys=%x virt=%p size=%x\n",
//	       __FUNCTION__, cfg->name, hose, start, hose->cfg_addr, size);

	/*********** I/O Window ***********/
	rio_get_bus_window(i, 1, &start, &size);
	page = (start - cfg->sgo_base) / RIO_SGO_PAGE_SIZE;
	count = (size + RIO_SGO_PAGE_SIZE - 1) / RIO_SGO_PAGE_SIZE;
	hose->io_base_phys = start;
	
	hose->io_base_virt = ioremap(start, size);
#ifdef RIO_WRITE_SGO_PAGES
	io_phys = rio_pci_sgo_write(cfg, page, count, SGO_TYPE_IO,
			hose->io_base_phys);
#endif
	cfg->io_top = (unsigned long)hose->io_base_virt + size;
	
	// pcibios_fixup_bus will convert the "phys" addr into a bus addr!
	rio_init_resource(&hose->io_resource, cfg->name,
		(unsigned long)hose->io_base_phys, size, IORESOURCE_IO);

	/*********** MEM Window ***********/
	rio_get_bus_window(i, 0, &start, &size);
#ifdef RIO_WRITE_SGO_PAGES
	page = (start - cfg->sgo_base) / RIO_SGO_PAGE_SIZE;
	count = size / RIO_SGO_PAGE_SIZE;
	mem_phys = rio_pci_sgo_write(cfg, page, count, SGO_TYPE_MEM | SGO_TYPE_D64, start);
#endif
	hose->pci_mem_offset = 0;

	cfg->mem_top = start + size;
	rio_init_resource(&hose->mem_resources[0], cfg->name,
		cfg->sgo_base, cfg->sgo_count * RIO_SGO_PAGE_SIZE, IORESOURCE_MEM);

//	hose->mem_resource_count = 1;
}

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,15)
extern int pci_assign_all_busses;
#else
extern int pci_assign_all_buses;
#endif

void __init rio_find_bridges(void)
{
	int i;
	rio_pci_cfg_t* cfg;

	/* This doesn't need to be earleaier (platform_init), so it is here in the file */
	rio_setup_pci_ptrs();
	rio_init_pci_slots();
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,15)
	pci_assign_all_busses=1;
#else
	pci_assign_all_buses=0;
#endif

	/* First, we need the CSR! */
	/* FIXME: Implement CSR sharing! */
	for (i = 0; i < RIO_PCI_BUSSES; i++) {
		CesResourceTableEntry_t const * res;
		cfg = &rio_pci_cfg[i];
		res = rio_get_bus_resource(i, 3);
		cfg->enabled = 0;
		if (!res)
			continue;
		if (!cesRT_IsMy(res))
			continue;
		cfg->csr = (rio_pci_csr_t*)(cesRT_GetAddress(res) - 0xf9009000 + RIO_PCI_REG_BASE);
		cfg->sgo_base = (cesRT_GetExtra(res) & 0xff) << 24;
		cfg->sgo_count = (cesRT_GetExtra(res) >> 8) & 0xff;
		cfg->enabled = 1;
//		printk("%s: Setting %s ->csr to %p\n", __FUNCTION__, cfg->name, cfg->csr);
	}

	printk("XPC-PCI bridges recognized:");

        /* before accessing the bridge, check whether the cap is there */


	if (lpci_cfg->enabled) {
		lpci_cfg->sys_slot = 1;
		printk(" LPCI");
	}
	if (xpci_cfg->enabled) {
		xpci_cfg->sys_slot = (in_le32(&xpci_cfg->csr->sts) & 0x02000000) ? 1:0;
		printk(" XPCI(%s)", xpci_cfg->sys_slot?"master":"slave");
	}
	if (cpci_cfg->enabled) {
		cpci_cfg->sys_slot = (in_le32(&cpci_cfg->csr->sts) & 0x02000000) ? 1:0;
		printk(" CPCI(%s)", cpci_cfg->sys_slot?"master":"slave");
	}
	if (epci_cfg->enabled) {
		epci_cfg->sys_slot = 1;
		printk(" EPCI");
	}
	printk("\n");

	/*
	 * Configure PCI buses present and enabled.
	 */
	for (i = 0; i < RIO_PCI_BUSSES; i++) {
		rio_init_pci_bus(i);
	}
}

#define PRINT_BUS_TREE
#ifdef PRINT_BUS_TREE
static inline void print_vertline(int depth, int line_mask) {
	int i;
	for (i = 0; i < depth; i++) {
		if (line_mask & (1 << i)) {
			printk("| ");
		} else {
			printk("  ");
		}
	}
	if (line_mask & 0x80000000) {
		if (line_mask & (1 << i)) {
			printk("|");
		} else {
			printk("`");
		}
	} else {
		if (line_mask & (1 << i)) {
			printk("| ");
		} else {
			printk("  ");
		}
	}
}

static void __init
rio_print_pci_tree_dev(struct pci_dev *dev, int depth, int line_mask)
{
	int i;
	struct resource *res;

	print_vertline(depth, line_mask | 0x80000000);
	printk("-%s IRQ 0x%x %04x:%04x\n", pci_name(dev), dev->irq, dev->vendor, dev->device);

	for (i = 0; i < PCI_ROM_RESOURCE; i++) {
		res = &dev->resource[i];
		if (res->flags & IORESOURCE_IO) {
			print_vertline(depth, line_mask);
			printk("             BAR %d: I/O 0x%08llx+0x%llx\n", i, (u64)res->start, (u64)(res->end - res->start + 1));
		} else if (res->flags & IORESOURCE_MEM) {
			print_vertline(depth, line_mask);
			printk("             BAR %d: MEM 0x%08llx+0x%llx\n", i, (u64)res->start, (u64)(res->end - res->start + 1));
		}
	}
}
static void __init
rio_print_pci_tree_bus(struct pci_bus *bus, const char *bus_name, int depth, int line_mask)
{
	struct pci_dev *dev;

	print_vertline(depth, line_mask | 0x80000000);
	printk("-+-%s BUS 0x%02x I/O 0x%08llx-0x%08llx MEM 0x%08llx-0x%08llx\n", bus_name, bus->number,
		(u64)bus->resource[0]->start, (u64)bus->resource[0]->end,
		(u64)bus->resource[1]->start, (u64)bus->resource[1]->end);
//	printk("-+-%s BUS 0x%02x\n", bus_name, bus->number);

	depth++;
	line_mask |= 1 << depth;

	/* Depth-First Search on bus tree */
	list_for_each_entry(dev, &bus->devices, bus_list) {
		if (dev->bus_list.next == &bus->devices)
			line_mask &= ~(1 << depth);
		if (dev->subordinate) {
			rio_print_pci_tree_bus(dev->subordinate, pci_name(dev), depth, line_mask);
		} else {
			rio_print_pci_tree_dev(dev, depth, line_mask);
		}
	}
}
int __init
rio_print_pci_tree(void)
{
	struct pci_bus *bus;
	int line_mask = 1;

	printk("+-PCI-TREE-------\n");
	list_for_each_entry(bus, &pci_root_buses, node) {
		struct pci_controller *hose = bus->sysdata;
		rio_pci_cfg_t *cfg = hose->arch_data;

		if (bus->node.next == &pci_root_buses)
			line_mask = 0;

		rio_print_pci_tree_bus(bus, cfg->name, 0, line_mask);
	}
	return 0;
}
device_initcall(rio_print_pci_tree);
#else
#define rio_print_pci_tree() do{}while(0)
#endif /* PRINT_BUS_TREE */

#if defined(CONFIG_HOTPLUG_CPCI) || defined(CONFIG_HOTPLUG_XPCI)
#include <linux/interrupt.h>
static irqreturn_t
rio_pci_enum(int irq, void *data, struct pt_regs *regs)
{
	rio_pci_cfg_t *cfg;
	int hs_csr = in_be32(&cfg->csr->hs_csr);
	printk("%cPCI ENUM:", cfg->name[0]);
	if (hs_csr & 0x00800000) {
		printk(" Insertion");
	}
	if (hs_csr & 0x00400000) {
		printk(" Extraction");
	}
	if (hs_csr & 0x00020000) {
		printk(" Signal");
	}
	printk("\n");
	if (cfg == xpci_cfg) {
		int sta = in_be32(&cfg->csr->sts);
		int i;
		for (i = 0; i < 3; i++) {
			if (sta & (0x20000 << i)) {
				printk("XPCI Slot %d is healthy\n", i + 1);
			} else {
				printk("XPCI Slot %d is not healthy\n", i + 1);
			}
		}
	} else if (cfg == cpci_cfg) {
		printk("FIXME: I dno't know how to find the added/removed device(s)\n");
	}
	return IRQ_HANDLED;
}
#endif

/* FIXME: should be renamed to rio_pci_late_init or rio_pci_init2 */
void __init rio_pci_request_resources(void) {
	/* Request the spaces used for CFG and I/O space from the memory root resource. 
	*/
	int i;

	for (i = 0; i < RIO_PCI_BUSSES; i++) {
		u32 start, size;
		char *name;
		rio_pci_cfg_t *cfg = &rio_pci_cfg[i];
		if (!cfg->enabled || !cfg->sys_slot)
			continue;

		rio_get_bus_window(i, 1, &start, &size);
		request_mem_region(start, size, "I/O Space");
		
		rio_get_bus_window(i, 2, &start, &size);
		request_mem_region(start, size, "CFG Space");

		name = kmalloc(32, GFP_KERNEL);
		sprintf(name, "%cPCI CSR", cfg->name[0]);
		request_mem_region((u32)cfg->csr, sizeof(rio_pci_csr_t), name);
	}
#ifdef CONFIG_HOTPLUG_XPCI
	if (xpci_cfg->enabled && xpci_cfg->sys_slot) {
		request_irq(RIO_ACT_XPCI_ENUM, rio_pci_enum, 0, "XPCI enum", xpci_cfg);
	}
#endif /* CONFIG_HOTPLUG_XPCI */
#ifdef CONFIG_HOTPLUG_CPCI
	if (cpci_cfg->enabled && cpci_cfg->sys_slot) {
		irq = RIO_ACT_CPCI_ENUM;
		request_irq(RIO_ACT_CPCI_ENUM, rio_pci_enum, 0, "CPCI enum", cpci_cfg);
	}
#endif /* CONFIG_HOTPLUG_CPCI */
}

/* RIO PCI OPS */

#define DPRINTK(x...)

int rio_pcibios_read_config(struct pci_bus *bus, unsigned int devfn, int offset,
			 int len, u32 *val)
{
	struct pci_controller *hose = bus->sysdata;
	rio_pci_cfg_t *cfg = hose->arch_data;
	int devno = PCI_SLOT(devfn);
	unsigned long cfg_data;


	if (hose->first_busno == bus->number &&
	    !pci_slot_is_enabled(cfg, devno)) {
		*val = ~0;
		return PCIBIOS_DEVICE_NOT_FOUND;
	}
	DPRINTK("read b: 0x%x, d: 0x%x, f: 0x%x, o: 0x%x, l: %d\n",
		bus->number, devfn>>3, devfn&7, offset, len);
	cfg_data = (unsigned long)hose->cfg_addr + (devfn << 8) + offset;
	if (bus->number != hose->first_busno)
		cfg_data |= bus->number << 16;
	switch (len) {
	case 1:
		*val = readb((volatile u8 __iomem *)cfg_data);
		break;
	case 2:
		*val = readw((volatile u16 __iomem *)cfg_data);
		break;
	default:
		*val = readl((volatile u32 __iomem *)cfg_data);
		break;
	}

	DPRINTK("v: 0x%x\n", *val);
	return PCIBIOS_SUCCESSFUL;
}

int
rio_pcibios_write_config(struct pci_bus *bus, unsigned int devfn, int offset,
			  int len, u32 val)
{
	struct pci_controller *hose = bus->sysdata;
	rio_pci_cfg_t *cfg = hose->arch_data;
	int devno = PCI_SLOT(devfn);
	unsigned long cfg_data;

	if (hose->first_busno == bus->number &&
	    !pci_slot_is_enabled(cfg, devno)) {
		return PCIBIOS_DEVICE_NOT_FOUND;
	}
	DPRINTK("write b: 0x%x, d: 0x%x, f: 0x%x, o: 0x%x, l: %d, v: 0x%x\n",
		bus->number, devfn>>3, devfn&7, offset, len, val);
	cfg_data = (unsigned long)hose->cfg_addr + (devfn << 8) + offset;
	if (bus->number != hose->first_busno)
		cfg_data |= bus->number << 16;
	switch (len) {
	case 1:
		writeb(val, (volatile u8 __iomem *)cfg_data); wmb();
		break;
	case 2:
		writew(val, (volatile u16 __iomem *)cfg_data); wmb();
		break;
	default:
		writel(val, (volatile u32 __iomem *)cfg_data); wmb();
/*		if (offset >= PCI_BASE_ADDRESS_0 && offset < (PCI_BASE_ADDRESS_0 * 6))
			printk("Write to dev %x.%x resource %d value %08lx\n", devno, devfn & 7, (offset - PCI_BASE_ADDRESS_0) / 4, val); */
		break;
	}

	return PCIBIOS_SUCCESSFUL;
}

static struct pci_ops rio_pci_ops = {
	rio_pcibios_read_config,
	rio_pcibios_write_config
};


#define DBG(x...) do{}while(0)

/* rounds-up start up to the next power-of-two of size */
static resource_size_t __init align_addr(resource_size_t start, resource_size_t size) {
	resource_size_t align = 0x1000;
	resource_size_t retval;
	
	while (align < size) {
		align <<= 1;
	}
	retval = (start + (align - 1)) & ~(align - 1);
	return retval;
}

void __init rio_pci_allocate_bus_resources(struct pci_bus *parent_bus, struct resource *parent_io_resource, struct resource *parent_mem_resource) {
	struct pci_bus *bus;
	struct pci_dev *dev;
	resource_size_t io_pos = parent_io_resource->start;
	resource_size_t mem_pos = parent_mem_resource->start;

	/* Depth-First Search on bus tree */
	list_for_each_entry(bus, &parent_bus->children, node) {
		struct resource *io_res = &bus->self->resource[0];
		struct resource *mem_res = &bus->self->resource[1];

		/* first, we give all of the remaining address space to the child bridge */
		io_res->start = io_pos;
		io_res->end = parent_io_resource->end;

		mem_res->start = mem_pos;
		mem_res->end = parent_mem_resource->end;

		rio_pci_allocate_bus_resources(bus, io_res, mem_res);

		update_bridge_resource(bus->self, io_res);
		update_bridge_resource(bus->self, mem_res);

		io_pos = io_res->end;
		mem_pos = mem_res->end;
	}
	list_for_each_entry(dev, &parent_bus->devices, bus_list) {
		int i;
		for (i = 0; i < PCI_ROM_RESOURCE; i++) {
			resource_size_t size;
			struct resource *res;
			res = &dev->resource[i];
			if (res->start == 0 && res->end == -1) {
				continue;
			}
			size = res->end - res->start + 1;
			if (res->flags & IORESOURCE_IO) {
				DBG("%s: moving device resource %d of %s from %08llx..%08llx\n",
				       __func__, i, pci_name(dev), (u64)res->start, (u64)res->end);
				io_pos = align_addr(io_pos, size);
				res->start = io_pos;
				res->end = io_pos + size - 1;
				io_pos = align_addr(io_pos + size, size);
				DBG("%s: moved device resource %d of %s to %08llx..%08llx\n",
				       __func__, i, pci_name(dev), (u64)res->start, (u64)res->end);

				pci_update_resource(dev, res, i);
			} else if (res->flags & IORESOURCE_MEM) {
				DBG("%s: moving device resource %d of %s from %08llx..%08llx\n",
				       __func__, i, pci_name(dev), (u64)res->start, (u64)res->end);
				mem_pos = align_addr(mem_pos, size);
				res->start = mem_pos;
				res->end = mem_pos + size - 1;
				mem_pos = align_addr(mem_pos + size, size);
				DBG("%s: moved device resource %d of %s to %08llx..%08llx\n",
				       __func__, i, pci_name(dev), (u64)res->start, (u64)res->end);

				pci_update_resource(dev, res, i);
			}
		}
	}
	parent_io_resource->end = io_pos - 1;
	parent_mem_resource->end = io_pos - 1;
}

static void __init rio_pcibios_fixup(void) {
	struct pci_controller *hose;
	struct pci_bus *bus;

	list_for_each_entry(bus, &pci_root_buses, node) {
		hose = bus->sysdata;
		if (!hose)
			continue;
		if (hose->arch_data == cpci_cfg) {
			struct resource *io_res  = bus->resource[0];
			struct resource *mem_res = bus->resource[1];
			resource_size_t io_end = io_res->end;
			resource_size_t mem_end = mem_res->end;

			DBG("%s: Fixing up CPCI bridge... IO=%08llx..%08llx MEM=%08llx..%08llx\n", __func__,
			       (u64)io_res->start, (u64)io_res->end,
			       (u64)mem_res->start, (u64)mem_res->end);

			rio_pci_allocate_bus_resources(bus, io_res, mem_res);

			DBG("%s: Fixed up CPCI bridge: IO=%08llx..%08llx MEM=%08llx..%08llx\n", __func__,
			       (u64)io_res->start, (u64)io_res->end,
			       (u64)mem_res->start, (u64)mem_res->end);

			hose->io_resource.end = io_end;
			hose->mem_resources[0].end = mem_end;
		}
	}
}
